# Copyright 2019 Virgil Dupras
#
# This software is licensed under the "GPLv3" License as described in the "LICENSE" file,
# which should be included with this package. The terms are also available at
# http://www.gnu.org/licenses/gpl-3.0.html

import datetime

from core.model.date import get_repeat_type_desc
from core.trans import trget, tr
from .column import Column
from ..model._ccore import amount_convert
from .table import GUITable, Row, rowattr, TableWithAmountMixin
from .util import splitted_splits

trcol = trget('columns')

class ScheduleTable(GUITable, TableWithAmountMixin):
    SAVENAME = 'ScheduleTable'
    COLUMNS = [
        Column('start_date', display=trcol('Start Date')),
        Column('stop_date', display=trcol('Stop Date')),
        Column('repeat_type', display=trcol('Repeat Type')),
        Column('interval', display=trcol('Interval')),
        Column('checkno', display=trcol('Check #'), visible=False, optional=True),
        Column('description', display=trcol('Description'), optional=True),
        Column('payee', display=trcol('Payee'), visible=False, optional=True),
        Column('from', display=trcol('From')),
        Column('to', display=trcol('To')),
        Column('amount', display=trcol('Amount')),
    ]

    def __init__(self, schedule_view):
        GUITable.__init__(self, document=schedule_view.document)
        self.mainwindow = schedule_view.mainwindow

    # --- Override
    def _update_selection(self):
        self.mainwindow.selected_schedules = self.selected_schedules

    def _fill(self):
        self._all_amounts_are_native = True
        for schedule in self.document.schedules:
            row = ScheduleTableRow(self, schedule)
            if not row.is_amount_native:
                self._all_amounts_are_native = False
            self.append(row)

    # --- Public
    def delete(self):
        self.document.delete_schedules(self.selected_schedules)
        self.mainwindow.revalidate()

    def edit(self):
        self.mainwindow.edit_item()

    # --- Properties
    @property
    def selected_schedules(self):
        return [row.schedule for row in self.selected_rows]

class ScheduleTableRow(Row):
    def __init__(self, table, schedule):
        Row.__init__(self, table)
        self.document = table.document
        self.schedule = schedule
        self.transaction = schedule.ref
        self.load()

    # --- Public
    def load(self):
        schedule = self.schedule
        txn = schedule.ref
        self._start_date = txn.date
        self._start_date_fmt = self.table.document.app.format_date(self._start_date)
        self._stop_date = schedule.stop_date
        if self._stop_date is not None:
            self._stop_date_fmt = self.table.document.app.format_date(self._stop_date)
        else:
            self._stop_date_fmt = ''
        self._repeat_type = get_repeat_type_desc(
            schedule.repeat_type, schedule.start_date)
        self._interval = str(schedule.repeat_every)
        self._description = txn.description
        self._payee = txn.payee
        self._checkno = txn.checkno
        froms, tos = splitted_splits(txn.splits)
        self._from_count = len(froms)
        self._to_count = len(tos)
        UNASSIGNED = tr('Unassigned') if len(froms) > 1 else ''
        self._from = ', '.join(s.account.name if s.account is not None else UNASSIGNED for s in froms)
        UNASSIGNED = tr('Unassigned') if len(tos) > 1 else ''
        self._to = ', '.join(s.account.name if s.account is not None else UNASSIGNED for s in tos)
        try:
            self._amount = sum(s.amount for s in tos)
        except ValueError: # currency coercing problem
            currency = self.document.default_currency
            self._amount = sum(amount_convert(s.amount, currency, s.transaction.date) for s in tos)
        self.is_amount_native = self.document.is_amount_native(self._amount)
        self._amount_fmt = self.document.format_amount(self._amount)

    def save(self):
        pass # read-only

    def sort_key_for_column(self, column_name):
        if column_name == 'stop_date' and self._stop_date is None:
            return datetime.date.min
        else:
            return Row.sort_key_for_column(self, column_name)

    # --- Properties
    start_date = rowattr('_start_date_fmt')
    stop_date = rowattr('_stop_date_fmt')
    repeat_type = rowattr('_repeat_type')
    interval = rowattr('_interval')
    description = rowattr('_description')
    payee = rowattr('_payee')
    checkno = rowattr('_checkno')
    from_ = rowattr('_from')
    to = rowattr('_to')
    amount = rowattr('_amount_fmt')
